@@ -15,24 +15,75 @@ also serve in other places.
 
 import os
 from os import path
-
-from filter import Filter
+from webassets.merge import BaseHunk
+from webassets.filter import Filter, freezedicts
+from webassets.utils import md5_constructor
 
 
 __all__ = ('FilesystemCache', 'MemoryCache', 'get_cache',)
 
 
+def make_hashable(data):
+    """Ensures ``data`` can be hashed().
+
+    Mostly needs to support dict. The other special types we use
+    as hash keys (Hunks, Filters) already have a proper hash() method.
+
+    See also ``make_md5``.
+
+    Note that we do not actually hash the data for the memory cache.
+    """
+    return freezedicts(data)
+
+
+def make_md5(data):
+    """Make a md5 hash based on``data``.
+
+    Specifically, this knows about ``Hunk`` objects, and makes sure
+    the actual content is hashed.
+
+    This is very conservative, and raises an exception if there are
+    data types that it does not explicitly support. This is because
+    we had in the past some debugging headaches with the cache not
+    working for this very reason.
+
+    MD5 is faster than sha, and we don't care so much about collisions.
+    We care enough however not to use hash().
+    """
+    def walk(obj):
+        if isinstance(obj, (tuple, list)):
+            for item in obj:
+                for d in walk(item): yield d
+        elif isinstance(obj, dict):
+            for k in sorted(obj.keys()):
+                for d in walk(k): yield d
+                for d in walk(obj[k]): yield d
+        elif isinstance(obj, BaseHunk):
+            yield obj.data()
+        elif isinstance(obj, Filter):
+            yield str(hash(obj))
+        elif isinstance(obj, (int, basestring)):
+            yield str(obj)
+        else:
+            raise ValueError('Cannot MD5 type %s' % type(obj))
+    md5 = md5_constructor()
+    for d in walk(data):
+        md5.update(d)
+    return md5.hexdigest()
+
+
 class BaseCache(object):
     """Abstract base class.
 
-    The cache key must be something that is supported by the
-    Python hash() function.
+    The cache key must be something that is supported by the Python hash()
+    function.
 
-    Since the cache is used for multiple purposes, all
-    webassets-internal code should always tag it's keys with
-    an id, like so:
+    Since the cache is used for multiple purposes, all webassets-internal code
+    should always tag it's keys with an id, like so:
 
         key = ("tag", actual_key)
+
+    One cache instance can only be used safely with a single Environment.
     """
 
     def get(self, key):
@@ -48,6 +99,10 @@ class MemoryCache(BaseCache):
 
     WARNING: Do NOT use this in a production environment, where you
     are likely going to have multiple processes serving the same app!
+
+    Note that the keys are used as-is, not passed through hash() (which is
+    a difference: http://stackoverflow.com/a/9022664/15677). However, the
+    reason we don't is because the original value is nicer to debug.
     """
 
     def __init__(self, capacity):
@@ -64,47 +119,11 @@ class MemoryCache(BaseCache):
                id(self) == id(other)
 
     def get(self, key):
+        key = make_hashable(key)
         return self.cache.get(key, None)
 
     def set(self, key, value):
-        self.cache[key] = value
-        try:
-            self.keys.remove(key)
-        except ValueError:
-            pass
-        self.keys.append(key)
-
-        # limit cache to the given capacity
-        to_delete = self.keys[0:max(0, len(self.keys)-self.capacity)]
-        self.keys = self.keys[len(to_delete):]
-        for item in to_delete:
-            del self.cache[key]
-
-
-class MemoryCache(BaseCache):
-    """Caches stuff in the process memory.
-
-    WARNING: Do NOT use this in a production environment, where you
-    are likely going to have multiple processes serving the same app!
-    """
-
-    def __init__(self, capacity):
-        self.capacity = capacity
-        self.keys = []
-        self.cache = {}
-
-    def __eq__(self, other):
-        """Return equality with the config values
-        that instantiate this instance.
-        """
-        return False == other or \
-               None == other or \
-               id(self) == id(other)
-
-    def get(self, key):
-        return self.cache.get(key, None)
-
-    def set(self, key, value):
+        key = make_hashable(key)
         self.cache[key] = value
         try:
             self.keys.remove(key)
@@ -135,7 +154,7 @@ class FilesystemCache(BaseCache):
                id(self) == id(other)
 
     def get(self, key):
-        filename = path.join(self.directory, '%s' % hash(key))
+        filename = path.join(self.directory, '%s' % make_md5(key))
         if not path.exists(filename):
             return None
         f = open(filename, 'rb')
@@ -145,7 +164,7 @@ class FilesystemCache(BaseCache):
             f.close()
 
     def set(self, key, data):
-        filename = path.join(self.directory, '%s' % hash(key))
+        filename = path.join(self.directory, '%s' % make_md5(key))
         f = open(filename, 'wb')
         try:
             f.write(data)
