@@ -7,7 +7,7 @@ This can be used as an alternative to an imperative setup.
 import os, sys
 from os import path
 import glob, fnmatch
-import importlib
+import types
 try:
     import yaml
 except ImportError:
@@ -41,19 +41,47 @@ class YAMLLoader(object):
             self.yaml = yaml
         self.file_or_filename = file_or_filename
 
+    def _yield_bundle_contents(self, data):
+        """Yield bundle contents from the given dict.
+
+        Each item yielded will be either a string representing a file path
+        or a bundle."""
+        contents = data.get('contents', [])
+        if isinstance(contents, basestring):
+            contents = contents,
+        for content in contents:
+            if isinstance(content, dict):
+                content = self._get_bundle(content)
+            yield content
+
+    def _get_bundle(self, data):
+        """Return a bundle initialised by the given dict."""
+        kwargs = dict(
+            filters=data.get('filters', None),
+            output=data.get('output', None),
+            debug=data.get('debug', None))
+        return Bundle(*list(self._yield_bundle_contents(data)), **kwargs)
+
     def _get_bundles(self, obj):
+        """Return a dict that keys bundle names to bundles."""
         bundles = {}
         for key, data in obj.iteritems():
             if data is None:
                 data = {}
-            contents = data.get('contents', [])
-            if isinstance(contents, basestring):
-                contents = [contents]
-            bundles[key] = Bundle(
-                filters=data.get('filters', None),
-                output=data.get('output', None),
-                *contents
-            )
+            bundles[key] = self._get_bundle(data)
+
+        # now we need to recurse through the bundles and get any that
+        # are included in each other.
+        for bundle_name, bundle in bundles.items():
+            # copy contents
+            contents = list(bundle.contents)
+            for i, item in enumerate(bundle.contents):
+                if item in bundles:
+                    contents[i] = bundles[item]
+            # cast back to a tuple
+            contents = tuple(contents)
+            if contents != bundle.contents:
+                bundle.contents = contents
         return bundles
 
     def _open(self):
@@ -98,6 +126,9 @@ class YAMLLoader(object):
             url: /media
             debug: True
             updater: timestamp
+            config:
+                compass_bin: /opt/compass
+                another_custom_config_value: foo
 
             bundles:
                 bundle-name:
@@ -122,10 +153,17 @@ class YAMLLoader(object):
             env = Environment(directory, obj['url'])
 
             # load environment settings
-            for setting in ('debug', 'cache', 'updater', 'expire',):
+            for setting in ('debug', 'cache', 'versions', 'url_expire',
+                            'auto_build',
+                            # TODO: The deprecated values; remove at some point
+                            'expire', 'updater'):
                 if setting in obj:
                     setattr(env, setting, obj[setting])
 
+            # load custom config options
+            if 'config' in obj:
+                env.config.update(obj['config'])
+
             # load bundles
             bundles = self._get_bundles(obj.get('bundles', {}))
             for name, bundle in bundles.iteritems():
@@ -142,25 +180,28 @@ class PythonLoader(object):
     """
 
     def __init__(self, module_name):
-        sys.path.insert(0, '')  # Ensure the current directory is on the path
-        try:
+        if isinstance(module_name, types.ModuleType):
+            self.module = module_name
+        else:
+            sys.path.insert(0, '')  # Ensure the current directory is on the path
             try:
-                self.module = import_module(module_name)
-            except ImportError, e:
-                raise LoaderError(e)
-        finally:
-            sys.path.pop(0)
+                try:
+                    self.module = import_module(module_name)
+                except ImportError, e:
+                    raise LoaderError(e)
+            finally:
+                sys.path.pop(0)
 
     def load_bundles(self):
         """Load ``Bundle`` objects defined in the Python module.
 
         Collects all bundles in the global namespace.
         """
-        bundles = []
+        bundles = {}
         for name in dir(self.module):
             value = getattr(self.module, name)
             if isinstance(value, Bundle):
-                bundles.append(value)
+                bundles[name] = value
         return bundles
 
     def load_environment(self):
